@@ -2,11 +2,11 @@
 '''Command line arguments parser
 '''
 
-import sys, traceback, getopt, types, textwrap, inspect
+import sys, traceback, getopt, types, textwrap, inspect, os
 from itertools import imap
 
 __all__ = ['command', 'dispatch']
-__version__ = '0.9.9'
+__version__ = '0.9.11'
 __author__ = 'Alexander Solovyov'
 __email__ = 'piranha@piranha.org.ua'
 
@@ -66,7 +66,7 @@ def command(options=None, usage=None, name=None, shortlist=False, hide=False):
             except StopIteration:
                 options_.append(('h', 'help', False, 'show help'))
 
-            argv = opts.pop('argv', None)
+            argv = opts.pop('argv', sys.argv[1:])
             if opts.pop('help', False):
                 return help_func()
 
@@ -74,21 +74,12 @@ def command(options=None, usage=None, name=None, shortlist=False, hide=False):
                 # no catcher here because this is call from Python
                 return call_cmd_regular(func, options_)(*args, **opts)
 
-            if argv is None:
-                argv = sys.argv[1:]
+            opts, args = catcher(lambda: parse(argv, options_), help_func)
 
-            try:
-                opts, args = catcher(lambda: parse(argv, options_), help_func)
-            except Abort:
-                return -1
-
-            try:
-                if opts.pop('help', False):
-                    return help_func()
-                return catcher(lambda: call_cmd(name_, func)(*args, **opts),
-                               help_func)
-            except Abort:
-                return -1
+            if opts.pop('help', False):
+                return help_func()
+            return catcher(lambda: call_cmd(name_, func)(*args, **opts),
+                           help_func)
 
         return inner
     return wrapper
@@ -126,15 +117,19 @@ def dispatch(args=None, cmdtable=None, globaloptions=None,
     cmdtable['help'] = (help_(cmdtable, globaloptions), [], '[TOPIC]')
     help_func = cmdtable['help'][0]
 
-    try:
-        name, func, args, kwargs = catcher(
-            lambda: _dispatch(args, cmdtable, globaloptions),
+    autocomplete(cmdtable, args, middleware)
+
+    name, func, args, kwargs = catcher(
+        lambda: _dispatch(args, cmdtable, globaloptions),
+        help_func)
+    if name == '_completion':       # skip middleware
+        return catcher(
+            lambda: call_cmd(name, func)(*args, **kwargs),
             help_func)
+    else:
         return catcher(
             lambda: call_cmd(name, middleware(func))(*args, **kwargs),
             help_func)
-    except Abort:
-        return -1
 
 # --------
 # Help
@@ -156,9 +151,9 @@ def help_(cmdtable, globalopts):
 
             for cmd, info in cmdtable.items():
                 if cmd.startswith('~'):
-                    continue # do not display hidden commands
+                    continue  # do not display hidden commands
                 if shortlist and not cmd.startswith('^'):
-                    continue # short help contains only marked commands
+                    continue  # short help contains only marked commands
                 cmd = cmd.lstrip('^~')
                 doc = info[0].__doc__ or '(no help text available)'
                 hlp[cmd] = doc.splitlines()[0].rstrip()
@@ -170,7 +165,7 @@ def help_(cmdtable, globalopts):
             write('\ncommands:\n\n')
             for cmd in hlplist:
                 doc = hlp[cmd]
-                if False: # verbose?
+                if False:  # verbose?
                     write(' %s:\n     %s\n' % (cmd.replace('|', ', '), doc))
                 else:
                     write(' %-*s  %s\n' % (maxlen, cmd.split('|', 1)[0],
@@ -223,18 +218,21 @@ def help_cmd(func, usage, options):
         --pid-file   name of file to write process ID to
     <BLANKLINE>
     '''
-    print '%s\n' % usage
+    write(usage + '\n')
     doc = func.__doc__
     if not doc:
         doc = '(no help text available)'
-    print '%s\n' % doc.strip()
+    write('\n' + doc.strip() + '\n\n')
     if options:
-        print ''.join(help_options(options))
+        write(''.join(help_options(options)))
 
 def help_options(options):
     yield 'options:\n\n'
     output = []
-    for short, name, default, desc in options:
+    for o in options:
+        short, name, default, desc = o[:4]
+        if hasattr(default, '__call__'):
+            default = default(None)
         default = default and ' (default: %s)' % default or ''
         output.append(('%2s%s' % (short and '-%s' % short,
                                   name and ' --%s' % name),
@@ -270,9 +268,11 @@ def parse(args, options):
 
     '''
     argmap, defmap, state = {}, {}, {}
-    shortlist, namelist = '', []
+    shortlist, namelist, funlist = '', [], []
 
-    for short, name, default, comment in options:
+    for o in options:
+        # might have the fifth completer element
+        short, name, default, comment = o[:4]
         if short and len(short) != 1:
             raise FOError('Short option should be only a single'
                           ' character: %s' % short)
@@ -287,7 +287,8 @@ def parse(args, options):
         # copy defaults to state
         if isinstance(default, list):
             state[pyname] = default[:]
-        elif callable(default):
+        elif hasattr(default, '__call__'):
+            funlist.append(pyname)
             state[pyname] = None
         else:
             state[pyname] = default
@@ -308,15 +309,17 @@ def parse(args, options):
         name = argmap[opt]
         t = type(defmap[name])
         if t is types.FunctionType:
+            del funlist[funlist.index(name)]
             state[name] = defmap[name](val)
-        elif t is types.IntType:
-            state[name] = int(val)
-        elif t is types.StringType:
-            state[name] = val
         elif t is types.ListType:
             state[name].append(val)
         elif t in (types.NoneType, types.BooleanType):
             state[name] = not defmap[name]
+        else:
+            state[name] = t(val)
+
+    for name in funlist:
+        state[name] = defmap[name](None)
 
     return state, args
 
@@ -361,6 +364,9 @@ def cmdparse(args, cmdtable, globalopts):
 
     return (cmd, cmd and info[0] or None, args, options)
 
+def aliases_(cmdtable_key):
+    return cmdtable_key.lstrip("^~").split("|")
+
 def findpossible(cmd, table):
     """
     Return cmd -> (aliases, command table entry)
@@ -368,7 +374,7 @@ def findpossible(cmd, table):
     """
     choice = {}
     for e in table.keys():
-        aliases = e.lstrip("^~").split("|")
+        aliases = aliases_(e)
         found = None
         if cmd in aliases:
             found = cmd
@@ -407,8 +413,9 @@ def guess_options(func):
     args, varargs, varkw, defaults = inspect.getargspec(func)
     for name, option in zip(args[-len(defaults):], defaults):
         try:
-            sname, default, hlp = option
-            yield (sname, name.replace('_', '-'), default, hlp)
+            sname, default, hlp = option[:3]
+            completer = option[3] if len(option) > 3 else None
+            yield (sname, name.replace('_', '-'), default, hlp, completer)
         except TypeError:
             pass
 
@@ -444,15 +451,6 @@ def catcher(target, help_func):
         help_func()
     except FOError, e:
         err('%s\n' % e)
-    except KeyboardInterrupt:
-        err('interrupted!\n')
-    except SystemExit:
-        raise
-    except:
-        err('unknown exception encountered')
-        raise
-
-    raise Abort
 
 def call_cmd(name, func):
     def inner(*args, **kwargs):
@@ -503,6 +501,92 @@ except ImportError:
             return wrapper
         return inner
 
+# --------
+# Autocomplete system
+# --------
+
+# Borrowed from PIP
+def autocomplete(cmdtable, args, middleware):
+    """Command and option completion.
+
+    Enable by sourcing one of the completion shell scripts (bash or zsh).
+    """
+
+    # Don't complete if user hasn't sourced bash_completion file.
+    if 'OPSTER_AUTO_COMPLETE' not in os.environ:
+        return
+    cwords = os.environ['COMP_WORDS'].split()[1:]
+    cword = int(os.environ['COMP_CWORD'])
+
+    try:
+        current = cwords[cword - 1]
+    except IndexError:
+        current = ''
+
+    commands = []
+    for k in cmdtable.keys():
+        commands += aliases_(k)
+
+    # command
+    if cword == 1:
+        print ' '.join(filter(lambda x: x.startswith(current), commands))
+
+    # command options
+    elif cwords[0] in commands:
+        idx = -2 if current else -1
+        options = []
+        aliases, (cmd, opts, usage) = findcmd(cwords[0], cmdtable)
+
+        for o in opts:
+            short, long, default, help = o[:4]
+            completer = o[4] if len(o) > 4 else None
+            short, long = '-%s' % short, '--%s' % long
+            options += [short, long]
+
+            if cwords[idx] in (short, long) and completer:
+                args = middleware(completer)(current)
+                print ' '.join(args),
+
+        print ' '.join((o for o in options if o.startswith(current)))
+
+    sys.exit(1)
+
+COMPLETIONS = {
+    'bash':
+        """
+# opster bash completion start
+_opster_completion()
+{
+    COMPREPLY=( $( COMP_WORDS="${COMP_WORDS[*]}" \\
+                   COMP_CWORD=$COMP_CWORD \\
+                   OPSTER_AUTO_COMPLETE=1 $1 ) )
+}
+complete -o default -F _opster_completion %s
+# opster bash completion end
+""",
+    'zsh':
+            """
+# opster zsh completion start
+function _opster_completion {
+  local words cword
+  read -Ac words
+  read -cn cword
+  reply=( $( COMP_WORDS="$words[*]" \\
+             COMP_CWORD=$(( cword-1 )) \\
+             OPSTER_AUTO_COMPLETE=1 $words[1] ) )
+}
+compctl -K _opster_completion %s
+# opster zsh completion end
+"""
+    }
+
+@command(name='_completion', hide=True)
+def completion(type=('t', 'bash', 'Completion type (bash or zsh)')):
+    """Outputs completion script for bash or zsh."""
+
+    prog_name = os.path.split(sys.argv[0])[1]
+    print COMPLETIONS[type] % prog_name
+
 # --------
 # Exceptions
 # --------
@@ -520,8 +604,5 @@ class UnknownCommand(CommandException):
 class ParseError(CommandException):
     'Raised on error in command line parsing'
 
-class Abort(CommandException):
-    'Abort execution'
-
 class FOError(CommandException):
     'Raised on trouble with opster configuration'
